@@ -46,6 +46,14 @@ struct imv_backend {
    * and src will point to an imv_source instance for the given data.
    */
   enum backend_result (*open_memory)(void *data, size_t len, struct imv_source **src);
+
+  /* Run initialization code before registering the the backend. If successful,
+   * BACKEND_SUCCESS is returned.
+   */
+  enum backend_result (*init)(void);
+
+  /* Uninitialize backend */
+  void (*uninit)(void);
 };
 
 #endif
