@@ -1,6 +1,7 @@
 #include "imv.h"
 
 #include <assert.h>
+#include <fcntl.h>
 #include <ctype.h>
 #include <errno.h>
 #include <getopt.h>
@@ -25,6 +26,7 @@
 #include "log.h"
 #include "navigator.h"
 #include "source.h"
+#include "sys/types.h"
 #include "viewport.h"
 #include "window.h"
 
@@ -116,7 +118,6 @@ struct imv {
   /* dirty state flags */
   bool need_redraw;
   bool need_rescale;
-  bool cache_invalidated;
 
   /* traverse sub-directories for more images */
   bool recursive_load;
@@ -176,6 +177,9 @@ struct imv {
   /* the user-specified format strings for the overlay and window title */
   char *title_text;
 
+  /* the user-specified string used as X11 res name or wayland app id */
+  char *app_id;
+
   /* imv subsystems */
   struct imv_binds *binds;
   struct imv_navigator *navigator;
@@ -223,6 +227,7 @@ static void render_window(struct imv *imv);
 static void update_env_vars(struct imv *imv);
 static size_t generate_env_text(struct imv *imv, char *buf, size_t len, const char *format);
 static size_t read_from_stdin(void **buffer);
+static void imv_backends_free(struct list *backends);
 
 /* Finds the next split between commands in a string (';'). Provides a pointer
  * to the next character after the delimiter as out, or a pointer to '\0' if
@@ -443,13 +448,12 @@ static void event_handler(void *data, const struct imv_event *e)
       break;
     case IMV_EVENT_RESIZE:
       {
-        const int ww = e->data.resize.width;
-        const int wh = e->data.resize.height;
-        const int bw = e->data.resize.buffer_width;
-        const int bh = e->data.resize.buffer_height;
-        const double scale = e->data.resize.scale;
-        imv_viewport_update(imv->view, ww, wh, bw, bh, imv->current_image, imv->scaling_mode);
-        imv_canvas_resize(imv->canvas, bw, bh, scale);
+        const int w = e->data.resize.buffer_width;
+        const int h = e->data.resize.buffer_height;
+        const double ui_scale = e->data.resize.scale;
+        imv_viewport_update(imv->view, w, h, imv->current_image, imv->scaling_mode);
+        imv_canvas_resize(imv->canvas, w, h);
+        imv_canvas_font(imv->canvas, imv->overlay.font.name, imv->overlay.font.size * ui_scale);
         break;
       }
     case IMV_EVENT_KEYBOARD:
@@ -469,6 +473,17 @@ static void event_handler(void *data, const struct imv_event *e)
             x, y, -e->data.mouse_scroll.dy);
       }
       break;
+    case IMV_EVENT_GESTURE_PINCH:
+      {
+        double x, y;
+        imv_window_get_mouse_position(imv->window, &x, &y);
+        imv_viewport_move(imv->view, e->data.gesture_pinch.dx,
+            e->data.gesture_pinch.dy, imv->current_image);
+        imv_viewport_scale_by(imv->view, imv->current_image, IMV_ZOOM_MOUSE,
+            x, y, e->data.gesture_pinch.scale);
+        imv_viewport_rotate_by(imv->view, e->data.gesture_pinch.rotation);
+      }
+      break;
     case IMV_EVENT_CUSTOM:
       consume_internal_event(imv, e->data.custom);
       break;
@@ -537,10 +552,6 @@ struct imv *imv_create(void)
   imv->commands = imv_commands_create();
   imv->console = imv_console_create();
   imv_console_set_command_callback(imv->console, &command_callback, imv);
-  imv->ipc = imv_ipc_create();
-  if (imv->ipc) {
-    imv_ipc_set_command_callback(imv->ipc, &command_callback, imv);
-  }
   imv->title_text = strdup(
       "imv - [${imv_current_index}/${imv_file_count}]"
       " [${imv_width}x${imv_height}] [${imv_scale}%]"
@@ -551,6 +562,7 @@ struct imv *imv_create(void)
       " [${imv_width}x${imv_height}] [${imv_scale}%]"
       " $imv_current_file [$imv_scaling_mode]"
   );
+  imv->app_id = strdup("imv");
   imv->overlay.text_color.r = 255;
   imv->overlay.text_color.g = 255;
   imv->overlay.text_color.b = 255;
@@ -637,6 +649,7 @@ void imv_free(struct imv *imv)
 {
   free(imv->overlay.font.name);
   free(imv->title_text);
+  free(imv->app_id);
   free(imv->overlay.text);
   imv_binds_free(imv->binds);
   imv_navigator_free(imv->navigator);
@@ -661,7 +674,7 @@ void imv_free(struct imv *imv)
     imv_window_free(imv->window);
   }
 
-  list_free(imv->backends);
+  imv_backends_free(imv->backends);
 
   list_free(imv->startup_commands);
 
@@ -670,7 +683,20 @@ void imv_free(struct imv *imv)
 
 void imv_install_backend(struct imv *imv, const struct imv_backend *backend)
 {
-  list_append(imv->backends, (void*)backend);
+  if (!backend->init || backend->init() == BACKEND_SUCCESS) {
+    list_append(imv->backends, (void*)backend);
+  }
+}
+
+static void imv_backends_free(struct list *backends)
+{
+  for (size_t i = 0; i < backends->len; ++i) {
+    struct imv_backend *backend = backends->items[i];
+    if (backend->uninit) {
+      backend->uninit();
+    }
+  }
+  list_free(backends);
 }
 
 static bool parse_bg(struct imv *imv, const char *bg)
@@ -731,6 +757,28 @@ static bool parse_upscaling_method(struct imv *imv, const char *method)
   return false;
 }
 
+static bool parse_window_title(struct imv *imv, const char *name)
+{
+  if (strcmp(name, "")) {
+    free(imv->title_text);
+    imv->title_text = strdup(name);
+    return true;
+  }
+
+  return false;
+}
+
+static bool parse_app_id(struct imv *imv, const char *id)
+{
+  if (strcmp(id, "")) {
+    free(imv->app_id);
+    imv->app_id = strdup(id);
+    return true;
+  }
+
+  return false;
+}
+
 static bool parse_initial_pan(struct imv *imv, const char *pan_params)
 {
   char *next_val;
@@ -743,6 +791,39 @@ static bool parse_initial_pan(struct imv *imv, const char *pan_params)
   return true;
 }
 
+static void set_cloexec(int fd)
+{
+  int flags = fcntl(fd, F_GETFD);
+  assert(flags != -1);
+  flags |= FD_CLOEXEC;
+  int rc = fcntl(fd, F_SETFD, flags);
+  assert(rc != -1);
+}
+
+static bool parse_initial_width(struct imv *imv, const char *width_value)
+{
+  char *endptr;
+  errno = 0;
+  int val_x = strtol(width_value, &endptr, 10);
+  if(errno != 0 || *endptr != '\0' || endptr == width_value) {
+    return false;
+  }
+  imv->initial_width = val_x;
+  return true;
+}
+
+static bool parse_initial_height(struct imv *imv, const char *height_value)
+{
+  char *endptr;
+  errno = 0;
+  int val_y = strtol(height_value, &endptr, 10);
+  if(errno != 0 || *endptr != '\0' || endptr == height_value) {
+    return false;
+  }
+  imv->initial_height = val_y;
+  return true;
+}
+
 static void *pipe_stdin(void *data)
 {
   int *fd = data;
@@ -770,6 +851,10 @@ static void *load_paths_from_stdin(void *data)
   char buf[PATH_MAX];
   while (fgets(buf, sizeof(buf), imv->stdin_pipe) != NULL) {
     size_t len = strlen(buf);
+    if (len == 0 || len >= PATH_MAX) {
+      imv_log(IMV_ERROR, "Failed to read path from stdin. Aborting.\n");
+      abort();
+    }
     if (buf[len-1] == '\n') {
       buf[--len] = 0;
     }
@@ -808,7 +893,7 @@ static void print_help(struct imv *imv)
   }
 
   puts("imv's full source code is published under the terms of the MIT\n"
-       "license, and can be found at https://github.com/eXeC64/imv\n"
+       "license, and can be found at https://sr.ht/~exec64/imv\n"
        "\n"
        "imv uses the inih library to parse ini files.\n"
        "See https://github.com/benhoyt/inih for details.\n"
@@ -823,7 +908,7 @@ bool imv_parse_args(struct imv *imv, int argc, char **argv)
   int o;
 
  /* TODO getopt_long */
-  while ((o = getopt(argc, argv, "frdxhvlu:s:n:b:t:c:")) != -1) {
+  while ((o = getopt(argc, argv, "frdxhvli:u:s:n:b:t:c:w:W:H:")) != -1) {
     switch(o) {
       case 'f': imv->start_fullscreen = true;                    break;
       case 'r': imv->recursive_load = true;                      break;
@@ -863,7 +948,21 @@ bool imv_parse_args(struct imv *imv, int argc, char **argv)
           return false;
         }
         break;
+      case 'W':
+        if(!parse_initial_width(imv, optarg)) {
+          imv_log(IMV_ERROR, "Invalid initial width value. Aborting.\n");
+          return false;
+        }
+        break;
+      case 'H':
+        if(!parse_initial_height(imv, optarg)) {
+          imv_log(IMV_ERROR, "Invalid initial height value. Aborting.\n");
+          return false;
+        }
+        break;
       case 'c': list_append(imv->startup_commands, optarg); break;
+      case 'w': parse_window_title(imv, optarg); break;
+      case 'i': parse_app_id(imv, optarg); break;
       case '?':
         imv_log(IMV_ERROR, "Unknown argument '%c'. Aborting.\n", optopt);
         return false;
@@ -915,6 +1014,11 @@ int imv_run(struct imv *imv)
   if (!setup_window(imv))
     return 1;
 
+  imv->ipc = imv_ipc_create();
+  if (imv->ipc) {
+    imv_ipc_set_command_callback(imv->ipc, &command_callback, imv);
+  }
+
   /* if loading paths from stdin, kick off a thread to do that - we'll receive
    * events back via internal events */
   int *stdin_pipe_fds = NULL;
@@ -930,6 +1034,9 @@ int imv_run(struct imv *imv)
       return 1;
     }
 
+    /* if a child process spawned by imv inherits the write part of the pipe,
+     * load_paths_from_stdin() will not exit until the child dies */
+    set_cloexec(stdin_pipe_fds[1]);
     imv->stdin_pipe = fdopen(stdin_pipe_fds[0], "re");
 
     if (pthread_create(&load_paths_thread, NULL, load_paths_from_stdin, imv)
@@ -939,22 +1046,56 @@ int imv_run(struct imv *imv)
   }
 
   if (imv->starting_path) {
-    ssize_t index = imv_navigator_find_path(imv->navigator, imv->starting_path);
-    if (index == -1) {
-      index = (int) strtol(imv->starting_path, NULL, 10);
-      index -= 1; /* input is 1-indexed, internally we're 0 indexed */
-      if (errno == EINVAL) {
-        index = -1;
+    if (imv->paths_from_stdin) {
+      int max_tries = 1000;
+      bool is_number = true;
+      for(size_t i = 0; i < strlen(imv->starting_path); ++i) {
+        if (!isdigit(imv->starting_path[i])) {
+          is_number = false;
+          break;
+        }
+      }
+      ssize_t index = -1;
+      if (is_number) {
+        index = strtol(imv->starting_path, NULL, 10);
+      }
+      bool cont = true;
+      while (cont) {
+        if (max_tries <= 0) {
+          cont = false;
+        }
+        imv_window_pump_events(imv->window, event_handler, imv);
+        if (index == -1) {
+          ssize_t img_index = imv_navigator_find_path(imv->navigator, imv->starting_path);
+          if(img_index != -1) {
+            imv_navigator_select_abs(imv->navigator, img_index);
+            cont = false;
+          }
+        } else {
+          if((ssize_t)imv_navigator_length(imv->navigator) >= index) {
+            imv_navigator_select_abs(imv->navigator, index);
+            cont = false;
+          }
+        }
+        max_tries -= 1;
+      }
+    } else {
+      ssize_t index = imv_navigator_find_path(imv->navigator, imv->starting_path);
+      if (index == -1) {
+        index = (int) strtol(imv->starting_path, NULL, 10);
+        index -= 1; /* input is 1-indexed, internally we're 0 indexed */
+        if (errno == EINVAL) {
+          index = -1;
+        }
+      }
+
+      if (index >= 0) {
+        imv_navigator_select_abs(imv->navigator, index);
+      } else {
+        imv_log(IMV_ERROR, "Invalid starting image: %s\n", imv->starting_path);
       }
     }
-
-    if (index >= 0) {
-      imv_navigator_select_abs(imv->navigator, index);
-    } else {
-      imv_log(IMV_ERROR, "Invalid starting image: %s\n", imv->starting_path);
-    }
   }
-
   /* Push any startup commands into the event queue */
   for (size_t i = 0; i < imv->startup_commands->len; ++i) {
     command_callback(imv->startup_commands->items[i], imv);
@@ -1002,6 +1143,11 @@ int imv_run(struct imv *imv)
               continue;
             }
 
+            if (!imv->stdin_image_data || !imv->stdin_image_data_len) {
+              /* Skip if image failed to load */
+              continue;
+            }
+
             result = backend->open_memory(imv->stdin_image_data,
                 imv->stdin_image_data_len, &new_source);
           } else {
@@ -1084,7 +1230,7 @@ int imv_run(struct imv *imv)
     }
 
     /* handle slideshow */
-    if (imv->slideshow.duration != 0.0) {
+    if (imv_viewport_is_playing(imv->view) && imv->slideshow.duration != 0.0) {
       double dt = current_time - last_time;
 
       imv->slideshow.elapsed += dt;
@@ -1160,7 +1306,8 @@ int imv_run(struct imv *imv)
 
 static bool setup_window(struct imv *imv)
 {
-  imv->window = imv_window_create(imv->initial_width, imv->initial_height, "imv");
+  imv->window = imv_window_create(imv->initial_width, imv->initial_height,
+                                  "imv", imv->app_id);
 
   if (!imv->window) {
     imv_log(IMV_ERROR, "Failed to create window\n");
@@ -1168,10 +1315,9 @@ static bool setup_window(struct imv *imv)
   }
 
   {
-    int ww, wh, bw, bh;
-    imv_window_get_size(imv->window, &ww, &wh);
-    imv_window_get_framebuffer_size(imv->window, &bw, &bh);
-    imv->view = imv_viewport_create(ww, wh, bw, bh);
+    int w, h;
+    imv_window_get_framebuffer_size(imv->window, &w, &h);
+    imv->view = imv_viewport_create(w, h);
   }
 
   if (imv->custom_start_pan) {
@@ -1182,9 +1328,9 @@ static bool setup_window(struct imv *imv)
   imv_window_set_fullscreen(imv->window, imv->start_fullscreen);
 
   {
-    int ww, wh;
-    imv_window_get_size(imv->window, &ww, &wh);
-    imv->canvas = imv_canvas_create(ww, wh);
+    int w, h;
+    imv_window_get_framebuffer_size(imv->window, &w, &h);
+    imv->canvas = imv_canvas_create(w, h);
     imv_canvas_font(imv->canvas, imv->overlay.font.name, imv->overlay.font.size);
   }
 
@@ -1267,9 +1413,6 @@ static void consume_internal_event(struct imv *imv, struct internal_event *event
 
 static void render_window(struct imv *imv)
 {
-  int ww, wh;
-  imv_window_get_size(imv->window, &ww, &wh);
-
   /* update window title */
   char title_text[1024];
   generate_env_text(imv, title_text, sizeof title_text, imv->title_text);
@@ -1285,10 +1428,6 @@ static void render_window(struct imv *imv)
         1.0);
     imv_canvas_fill(imv->canvas);
     imv_canvas_draw(imv->canvas);
-  } else {
-    /* chequered background */
-    imv_canvas_fill_checkers(imv->canvas, 16);
-    imv_canvas_draw(imv->canvas);
   }
 
   /* draw our actual image */
@@ -1300,59 +1439,75 @@ static void render_window(struct imv *imv)
     imv_viewport_get_scale(imv->view, &scale);
     imv_viewport_get_rotation(imv->view, &rotation);
     imv_viewport_get_mirrored(imv->view, &mirrored);
+    if (imv->background.type == BACKGROUND_CHEQUERED) {
+      imv_canvas_fill_checkers(imv->canvas, imv->current_image,
+                               x, y, scale, rotation, mirrored);
+    }
     imv_canvas_draw_image(imv->canvas, imv->current_image,
                           x, y, scale, rotation, mirrored,
-                          imv->upscaling_method, imv->cache_invalidated);
+                          imv->upscaling_method);
   }
 
   imv_canvas_clear(imv->canvas);
 
+  int w, h;
+  imv_window_get_framebuffer_size(imv->window, &w, &h);
+  int ui_scale = imv_window_get_scale(imv->window);
+
   /* if the overlay needs to be drawn, draw that too */
   if (imv->overlay.enabled) {
-    const int height = imv->overlay.font.size * 1.2;
+    char overlay_text[1024];
+    generate_env_text(imv, overlay_text, sizeof overlay_text, imv->overlay.text);
+    PangoLayout *layout = imv_canvas_make_layout(imv->canvas, overlay_text);
+
+    int width, height;
+    pango_layout_get_pixel_size(layout, &width, &height);
+
+    int y = 0;
+    const int bottom_offset = 5 * ui_scale;
+    if (imv->overlay.position_at_bottom)
+    {
+      y = h - height - bottom_offset;
+    }
+
     imv_canvas_color(imv->canvas,
         imv->overlay.background_color.r / 255.f,
         imv->overlay.background_color.g / 255.f,
         imv->overlay.background_color.b / 255.f,
         imv->overlay.background_alpha / 255.f);
-    int y = 0 ;
-    const int bottom_offset = 5;
-    if (imv->overlay.position_at_bottom)
-    {
-      y = wh - height - bottom_offset;
-    }
-    imv_canvas_fill_rectangle(imv->canvas, 0, y, ww, height + bottom_offset);
+    imv_canvas_fill_rectangle(imv->canvas, 0, y, width, height + bottom_offset);
+
     imv_canvas_color(imv->canvas,
         imv->overlay.text_color.r / 255.f,
         imv->overlay.text_color.g / 255.f,
         imv->overlay.text_color.b / 255.f,
         imv->overlay.text_alpha / 255.f);
-    char overlay_text[1024];
-    generate_env_text(imv, overlay_text, sizeof overlay_text, imv->overlay.text);
-    imv_canvas_printf(imv->canvas, 0, y, "%s", overlay_text);
+    imv_canvas_show_layout(imv->canvas, 0, y, layout);
+
+    g_object_unref(layout);
   }
 
   /* draw command entry bar if needed */
   if (imv_console_prompt(imv->console)) {
-    const int bottom_offset = 5;
-    const int height = imv->overlay.font.size * 1.2;
+    const int bottom_offset = 5 * ui_scale;
+    const int height = imv->overlay.font.size * ui_scale * 1.2;
     imv_canvas_color(imv->canvas, 0, 0, 0, 0.75);
-    imv_canvas_fill_rectangle(imv->canvas, 0, wh - height - bottom_offset,
-        ww, height + bottom_offset);
+    imv_canvas_fill_rectangle(imv->canvas, 0, h - height - bottom_offset,
+        w, height + bottom_offset);
     imv_canvas_color(imv->canvas, 1, 1, 1, 1);
 
     int x = 0;
     /* draw pre-cursor text */
-    x += imv_canvas_printf(imv->canvas, x, wh - height - bottom_offset,
+    x += imv_canvas_printf(imv->canvas, x, h - height - bottom_offset,
         ":%.*s",
         imv_console_prompt_cursor(imv->console),
         imv_console_prompt(imv->console));
     /* draw the cursor */
     imv_canvas_color(imv->canvas, 1, 1, 1, 0.5);
-    imv_canvas_printf(imv->canvas, x, wh - height - bottom_offset, "\u2588");
+    imv_canvas_printf(imv->canvas, x, h - height - bottom_offset, "\u2588");
     /* any any remaining text on top of the cursor */
     imv_canvas_color(imv->canvas, 1, 1, 1, 1);
-    imv_canvas_printf(imv->canvas, x, wh - height - bottom_offset, "%s",
+    imv_canvas_printf(imv->canvas, x, h - height - bottom_offset, "%s",
         imv_console_prompt(imv->console) + imv_console_prompt_cursor(imv->console));
   }
 
@@ -1360,7 +1515,6 @@ static void render_window(struct imv *imv)
 
   /* redraw complete, unset the flag */
   imv->need_redraw = false;
-  imv->cache_invalidated = false;
 }
 
 static char *get_config_path(void)
@@ -1657,8 +1811,13 @@ static void command_zoom(struct list *args, const char *argstr, void *data)
     if (!strcmp(str, "actual")) {
       imv_viewport_scale_to_actual(imv->view, imv->current_image);
     } else {
-      long int amount = strtol(args->items[1], NULL, 10);
-      imv_viewport_zoom(imv->view, imv->current_image, IMV_ZOOM_KEYBOARD, 0, 0, amount);
+      char *endptr;
+      long int amount = strtol(args->items[1], &endptr, 10);
+      if (*endptr == '%') {
+        imv_viewport_scale_by(imv->view, imv->current_image, IMV_ZOOM_KEYBOARD, 0, 0, amount/100.0);
+      } else if (*endptr == '\0') {
+        imv_viewport_zoom(imv->view, imv->current_image, IMV_ZOOM_KEYBOARD, 0, 0, amount);
+      }
     }
   }
 }
@@ -1693,26 +1852,26 @@ static void command_flip(struct list *args, const char *argstr, void *data)
 
 static void command_open(struct list *args, const char *argstr, void *data)
 {
-  (void)argstr;
+  (void)args;
+  while(*argstr == ' ') {
+    argstr++;
+  }
   struct imv *imv = data;
   bool recursive = imv->recursive_load;
 
   update_env_vars(imv);
-  for (size_t i = 1; i < args->len; ++i) {
+  /* allow -r arg to specify recursive */
+  if (memcmp(argstr, "-r", 2) == 0) {
+    argstr += 2;
+    recursive = true;
+  }
 
-    /* allow -r arg to specify recursive */
-    if (i == 1 && !strcmp(args->items[i], "-r")) {
-      recursive = true;
-      continue;
-    }
-
-    wordexp_t word;
-    if (wordexp(args->items[i], &word, 0) == 0) {
-      for (size_t j = 0; j < word.we_wordc; ++j) {
-        imv_navigator_add(imv->navigator, word.we_wordv[j], recursive);
-      }
-      wordfree(&word);
+  wordexp_t word;
+  if (wordexp(argstr, &word, 0) == 0) {
+    for (size_t i = 0; i < word.we_wordc; ++i) {
+      imv_navigator_add(imv->navigator, word.we_wordv[i], recursive);
     }
+    wordfree(&word);
   }
 }
 
@@ -1847,7 +2006,6 @@ static void command_set_upscaling_method(struct list *args, const char *argstr,
   }
 
   imv->need_redraw = true;
-  imv->cache_invalidated = true;
 }
 
 static void command_set_slideshow_duration(struct list *args, const char *argstr, void *data)
@@ -1945,14 +2103,16 @@ static size_t generate_env_text(struct imv *imv, char *buf, size_t buf_len, cons
 
   size_t len = 0;
   wordexp_t word;
+  setenv("IFS", "", 1);
   if (wordexp(format, &word, 0) == 0) {
     for (size_t i = 0; i < word.we_wordc; ++i) {
-      len += snprintf(buf + len, buf_len - len, "%s ", word.we_wordv[i]);
+      len += snprintf(buf + len, buf_len - len, (i ? " %s" : "%s"), word.we_wordv[i]);
     }
     wordfree(&word);
   } else {
     len += snprintf(buf, buf_len, "error expanding text");
   }
+  unsetenv("IFS");
 
   return len;
 }
@@ -1962,27 +2122,41 @@ static size_t read_from_stdin(void **buffer)
   size_t len = 0;
   ssize_t r;
   size_t step = 4096; /* Arbitrary value of 4 KiB */
-  void *p;
+  void *new_buf;
 
-  errno = 0; /* clear errno */
+  errno = 0;
+  *buffer = NULL;
 
-  for (*buffer = NULL; (*buffer = realloc((p = *buffer), len + step));
-      len += (size_t)r) {
-    if ((r = read(STDIN_FILENO, (uint8_t *)*buffer + len, step)) == -1) {
-      perror(NULL);
+  while (1) {
+    new_buf = realloc(*buffer, len + step);
+    if (new_buf) {
+      *buffer = new_buf;
+    } else {
+      /* Failed to extend buffer */
+      int save = errno;
+      free(*buffer);
+      errno = save;
+      *buffer = NULL;
+      len = 0;
       break;
-    } else if (r == 0) {
+    }
+
+    r = read(STDIN_FILENO, (uint8_t *)*buffer + len, step);
+    if (r > 0) {
+      len += (size_t)r;
+    } else {
+      if (r < 0) {
+        /* Read error */
+        int save = errno;
+        perror(NULL);
+        free(*buffer);
+        errno = save;
+        len = 0;
+      }
       break;
     }
   }
 
-  /* realloc(3) leaves old buffer allocated in case of error */
-  if (*buffer == NULL && p != NULL) {
-    int save = errno;
-    free(p);
-    errno = save;
-    len = 0;
-  }
   return len;
 }
 
